import sys

from loguru import logger

from deciphon_sched.settings import Settings


class Logger:
    def __init__(self, settings: Settings):
        logger.remove()
        logger.add(sys.stderr, level=settings.log_level.value.upper())

    @property
    def handler(self):
        return logger



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Optional

import fastapi
from deciphon_core.schema import (
    DB_NAME_PATTERN,
    HMM_NAME_PATTERN,
    NAME_MAX_LENGTH,
    SNAP_NAME_PATTERN,
    DBName,
    Gencode,
    HMMName,
    SnapName,
)
from deciphon_snap.prod import Prod
from pydantic import BaseModel


class JobType(Enum):
    hmm = "hmm"
    scan = "scan"


class JobState(Enum):
    pend = "pend"
    run = "run"
    done = "done"
    fail = "fail"


class JobRead(BaseModel):
    id: int
    type: JobType
    state: JobState
    progress: int
    error: str
    submission: datetime
    exec_started: Optional[datetime]
    exec_ended: Optional[datetime]


class JobUpdate(BaseModel):
    state: JobState
    progress: int
    error: str


class HMMFile(HMMName):
    @property
    def db_name(self):
        return DBName(name=self.name[:-4] + ".dcp")

    @property
    def path_type(self):
        return fastapi.Path(
            title="HMM file",
            pattern=HMM_NAME_PATTERN,
            max_length=NAME_MAX_LENGTH,
        )


class DBFile(DBName):
    gencode: Gencode
    epsilon: float

    @property
    def hmm_name(self):
        return HMMName(name=self.name[:-4] + ".hmm")

    @property
    def path_type(self):
        return fastapi.Path(
            title="DB file",
            pattern=DB_NAME_PATTERN,
            max_length=NAME_MAX_LENGTH,
        )


class SnapFile(SnapName):
    @property
    def path_type(self):
        return fastapi.Path(
            title="Snap file",
            pattern=SNAP_NAME_PATTERN,
            max_length=NAME_MAX_LENGTH,
        )


class PressRequest(BaseModel):
    job_id: int
    hmm: HMMFile
    db: DBFile

    @classmethod
    def create(cls, job_id: int, hmm: HMMFile, gencode: Gencode, epsilon: float):
        db = DBFile(name=hmm.db_name.name, gencode=gencode, epsilon=epsilon)
        return cls(job_id=job_id, hmm=hmm, db=db)


class ScanRequest(BaseModel):
    id: int
    job_id: int
    hmm: HMMFile
    db: DBFile
    multi_hits: bool
    hmmer3_compat: bool
    seqs: list[SeqRead]

    @classmethod
    def create(cls, scan: ScanRead):
        return cls(
            id=scan.id,
            job_id=scan.job.id,
            hmm=HMMFile(name=scan.db.hmm.file.name),
            db=scan.db.file,
            multi_hits=scan.multi_hits,
            hmmer3_compat=scan.hmmer3_compat,
            seqs=scan.seqs,
        )


class HMMRead(BaseModel):
    id: int
    job: JobRead
    file: HMMFile


class DBCreate(BaseModel):
    file: DBFile


class DBRead(BaseModel):
    id: int
    hmm: HMMRead
    file: DBFile


class SeqCreate(BaseModel):
    name: str
    data: str


class SeqRead(BaseModel):
    id: int
    name: str
    data: str


class SnapRead(BaseModel):
    id: int
    size: int


class ScanCreate(BaseModel):
    db_id: int
    multi_hits: bool
    hmmer3_compat: bool
    seqs: list[SeqCreate]


class ScanRead(BaseModel):
    id: int
    job: JobRead
    db: DBRead
    multi_hits: bool
    hmmer3_compat: bool
    seqs: list[SeqRead]


class ProdRead(BaseModel):
    seq_id: int
    profile: str
    abc: str
    alt: float
    null: float
    evalue: float

    @classmethod
    def make(cls, x: Prod):
        return cls(
            seq_id=x.seq_id,
            profile=x.profile,
            abc=x.abc,
            alt=x.alt,
            null=x.null,
            evalue=x.evalue,
        )



from sqlalchemy import MetaData, create_engine
from sqlalchemy.orm import Session
from deciphon_sched.logger import Logger

from deciphon_sched.settings import Settings


class Database:
    def __init__(self, settings: Settings, logger: Logger):
        self._engine = create_engine(settings.database_url.unicode_string())
        self._logger = logger

    def create_tables(self, metadata: MetaData):
        self._logger.handler.debug("creating database tables")
        metadata.create_all(self._engine)

    def create_session(self):
        return Session(self._engine)

    def metadata(self):
        x = MetaData()
        x.reflect(self._engine)
        return x

    def dispose(self):
        self._logger.handler.debug("disposing database engine")
        self._engine.dispose()



from __future__ import annotations

import paho.mqtt.client
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_container_is_ready
from testcontainers.minio import MinioContainer


def mqtt_server():
    with MosquittoContainer() as x:
        yield {
            "host": x.get_container_host_ip(),
            "port": int(x.get_exposed_port(x.port)),
        }


def s3_server():
    with MinioContainer() as x:
        yield {
            "client": x.get_client(),
            "access_key": x.get_config()["access_key"],
            "secret_key": x.get_config()["secret_key"],
            "url": "http://" + x.get_config()["endpoint"],
        }


def s3_cleanup(client):
    for bucket in client.list_buckets():
        for x in client.list_objects(bucket.name):
            client.remove_object(bucket.name, x.object_name)
        client.remove_bucket(bucket.name)


class MosquittoContainer(DockerContainer):
    def __init__(self, image="eclipse-mosquitto:2", port=1883):
        super(MosquittoContainer, self).__init__(image)
        self.port = port
        self.with_exposed_ports(self.port)

    @wait_container_is_ready(ConnectionError)
    def _healthcheck(self):
        host = self.get_container_host_ip()
        port = int(self.get_exposed_port(self.port))
        paho.mqtt.client.Client().connect(host, port, 1)

    def start(self):
        self.with_command("mosquitto -c /mosquitto-no-auth.conf")
        super().start()
        self._healthcheck()
        return self



from fastapi import APIRouter, Request
from starlette.status import HTTP_200_OK

from deciphon_sched.database import Database
from deciphon_sched.errors import NotFoundInDatabaseError
from deciphon_sched.sched.models import Job
from deciphon_sched.sched.schemas import JobRead, JobState, JobUpdate


router = APIRouter()


@router.get("/jobs", status_code=HTTP_200_OK)
async def read_jobs(request: Request) -> list[JobRead]:
    database: Database = request.app.state.database
    with database.create_session() as session:
        return [x.read_model() for x in Job.get_all(session)]


@router.get("/jobs/{job_id}", status_code=HTTP_200_OK)
async def read_job(request: Request, job_id: int) -> JobRead:
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = Job.get_by_id(session, job_id)
        if x is None:
            raise NotFoundInDatabaseError("Job")
        return x.read_model()


@router.patch("/jobs/{job_id}", status_code=HTTP_200_OK)
async def update_job_state(request: Request, job_id: int, job: JobUpdate) -> JobRead:
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = Job.get_by_id(session, job_id)
        if x is None:
            raise NotFoundInDatabaseError("Job")

        if job.state == JobState.pend:
            x.set_pend()
        if job.state == JobState.run:
            x.set_run(job.progress)
        elif job.state == JobState.done:
            x.set_done()
        elif job.state == JobState.fail:
            x.set_fail(job.error)
        else:
            assert False

        session.commit()
        return x.read_model()
